import path from 'path'
import { omit } from 'ramda';

const pathSep = path.sep

export function defaultFactory() {
  const fileToIdMap = new Map();
  let nextId = 0;
  return (modulePath: string) => {
    let id = fileToIdMap.get(modulePath);
    if (typeof id !== 'number') {
      id = nextId++;
      fileToIdMap.set(modulePath, id);
    }
    return id;
  };
}

type Options = {
  projectRoot: string
  filterModules: Set<string>
  SDKAlias: { [aliasName: string]: string[] }
}


const caches = new Map<string, string>()

const getId = (modulePath: string, options: Omit<Options, 'filterModules'>) => {
  const {
    projectRoot,
    SDKAlias,
  } = options
  const genSDKReg = (SDKNames: string[]) => {
    const uniformNames = SDKNames.map(x => x.replace('/', '_'))
    return new RegExp(`node_modules_(${uniformNames.join('|')})`, 'ig')
  }

  //RN JS 全部以 RN_ 开头
  if (modulePath.includes('node_modules' + pathSep + 'react-native' + pathSep + 'Libraries' + pathSep)) {
    return `RN_${path.basename(modulePath, '.js')}`
  }


  // 如果模块不在 projectRoot 内, 比如 metro-runtime/src/polyfills/requires.js
  // id 变成了 _Users_simon_Developer_huya_hyext_node-pack_node_modules_metro-runtime_src_polyfills_require
  // 为了能正确过滤basebundle中已有的模块(node_modules_metro-runtime_src_polyfills_require), 需要进行处理
  let name = [projectRoot, process.env.MATRIX_SHARE_PROJECT_RN_ROOT!, process.env.MATRIX_PACKER_ROOT!]
    .filter(Boolean)
    .reduce((acc, cur) => {
      return acc.replace(`${cur}/`, '')
    }, modulePath)

  // remove .js extendtion
  name = name.replace(/.js$/, '')

  // 路径分隔符换成 _
  const regExp = pathSep == '\\' ? new RegExp('\\\\', 'g') : new RegExp(pathSep, 'g')
  name = name.replace(regExp, '_')

  // 把业务/base jsbundle 里的非 node_modules 里的模块 id 都带上根目录名
  if (modulePath.includes(projectRoot) && !modulePath.includes('node_modules')) {
    name = path.basename(projectRoot) + '_' + name;
  }

  Object.entries(SDKAlias).forEach(([alias, SDKNames]) => {
    if (SDKNames.length <= 0) return

    const SDKReg = genSDKReg(SDKNames)
    if (SDKReg.test(name)) {
      name = name.replace(SDKReg, alias)
    }
  })

  return name;
}

export function createModuleIdHOF(opts: Options) {
  return function createModuleIdFactory() { // 既然称之为 Factory, 表明这个函数会被多次调用
    return (modulePath: string) => {
      const cacheId = caches.get(modulePath)
      if (cacheId) return cacheId

      let moduleId = getId(modulePath, omit(['filterModules'], opts))

      // 如果是非baseBundle中的包, 那么加上一些额外的字符来避免以下情况:
      // 1. baseBundle加了一些包, 但是没有在同步到模块清单中
      // 2. 如果baseBundle/busiBundle 都用同一个模块名称, 那么可能会导致"重复定义"的错误
      // 所以, 非baseBundle模块清单中的模块, 在命名上做一些个性化的区分
      if (opts.filterModules.size > 0 && !opts.filterModules.has(moduleId)) {
        moduleId = `${moduleId}_#biz`
      }

      caches.set(modulePath, moduleId)

      return moduleId
    }
  }
}
