import path from 'path'
// @ts-ignore untyped metro-cache
import { FileStore } from 'metro-cache'

// @ts-ignore
import { ConfigT, mergeConfig } from 'metro-config'
// @ts-ignore
import { getDefaultValues } from 'metro-config/src/defaults'
import { getDefaultConfig as getOfficialCliConfig } from '@react-native-community/cli-plugin-metro'

import { defaultFactory, createModuleIdHOF } from './createModuleIdHOF'
import resolveReactNativePath from './resolveReactNativePath'
import filterModuleHOF from './filterModuleHOF'
import { uniq } from 'ramda'
import { ExtMetroConfig } from './configTypes'

interface BaseOptions {
  projectRoot: string
  extMetroConfig: Partial<ExtMetroConfig>
  projectConfig?: {
    cacheVersion?: number
    [key: string]: any
  }
}

interface DevOptions extends BaseOptions {
  mode: 'dev',
  port?: number
}

interface ProdOptions extends BaseOptions {
  mode: 'production'

  /*
   * SDK 的npm包名称, 包含 org 前缀(如有), 如 ['@hyext/hyext-rn-sdk'], ['hyext-rn-sdk']
   * 业务代码对 sdk 的引用将会被替换成 open-sdk(open-sdk 在app内预置了, 可直接使用)
   * 如 node_modules_@hyext_hyext-rn-sdk 将会变替换成 open-sdk
   */
  // SDKAlias?: { [aliasName: string]: string[] }
  // filterModules?: Set<string>
  // filterModulesVersion?: Set<string>
  // extraModulesRunFirst?: string[]
  // extraPolyfills?: string[]
}

export const skipModulePlaceHolder = {
  dependencies: [],
  output: [],
  getSource: () => Buffer.from(''),
}

// TODO: 搜索 ext-sdk 所在文件夹
export function getExtMetroConfig(options: DevOptions | ProdOptions): ExtMetroConfig {
  const {
    projectRoot,
    mode,
    projectConfig,
    extMetroConfig: {
      filterModules = new Set(),
      filterModulesVersion = new Set(),
      SDKAlias = {},
      extraPolyfills = [],
      extraModulesRunFirst = [],
    },
  } = options

  // metro 默认值: metro-config/src/defaults/index.js
  const officialDefaultConfig = getDefaultValues(projectRoot)
  const officialCliConfig = getOfficialCliConfig({
    platforms: {},
    reactNativePath: resolveReactNativePath([process.env.MATRIX_SHARE_PROJECT_RN_ROOT!, projectRoot].filter(Boolean)),
  } as any)
  const officialConfig: ConfigT = mergeConfig(officialDefaultConfig, officialCliConfig)

  // 所以的依赖必须要归到 watchFolder 中去, 否则找不到相关的文件, 或者文件缺失md5
  const extraWatchFolder = [
    'metro-runtime/src/modules/asyncRequire', // metro-runtime could come from here
    '@react-native-community/cli-plugin-metro', // or come from here
  ]
    .map(xs => require.resolve(xs))
    .map(xs => require.resolve('metro-runtime/src/modules/asyncRequire', { paths: [xs] })) // watch asyncRequire 所在的文件夹
    .map(xs => path.dirname(xs))
  // watch matrix-config 所在的文件夹
  extraWatchFolder.push(path.resolve(__dirname))
  // watch metro-runtime/src 文件夹
  extraWatchFolder.push(
    path.join(path.dirname(require.resolve('metro-runtime/package.json')), 'src')
  )
  extraWatchFolder.push(
    ...extraModulesRunFirst.map(xs => path.dirname(xs)),
    ...extraPolyfills.map(xs => path.dirname(xs)),
  )

  /*
   * 开发环境下, 这个值设置为true, 否则会遇到问题:
   * https://github.com/facebook/react-native/issues/28801
   */
  let inlineRequires = true
  let watchFolders = uniq([projectRoot, ...extraWatchFolder])
  let nodeModulesPaths = [path.join(projectRoot, 'node_modules')]
  let assetRegistryPath = officialConfig.transformer.assetRegistryPath
  let useWatchman = true
  let createModuleIdFactory = defaultFactory

  if (mode !== 'dev') {
    inlineRequires = false
    assetRegistryPath = path.resolve(path.join(__dirname, 'assetRegistry.js'))
    useWatchman = false
    createModuleIdFactory = createModuleIdHOF({ projectRoot, SDKAlias, filterModules })

    if (process.env.MATRIX_SHARE_PROJECT_RN_ROOT) {
      const shareModules = path.join(process.env.MATRIX_SHARE_PROJECT_RN_ROOT, 'node_modules')
      // 大部分包(react/react-dom/react-native/sdk)其实都在 projectRoot/node_modules 这一层,
      // 所以先搜索这个位置比较高效
      nodeModulesPaths = [shareModules, ...nodeModulesPaths]
      watchFolders.push(shareModules)
    }
  }

  const CACHE_PATH = path.join(
    process.env.MATRIX_CACHE_ROOT || path.join(projectRoot, 'node_modules', '.huya-ext-cache'),
    'metro',
  )

  const hyextConfig = {
    resolver: {
      sourceExts: [...officialConfig.resolver.sourceExts, 'jsx', 'hycss'],
      platforms: ['ios', 'android', 'native', 'web'],
      nodeModulesPaths,
      blockList: [
        ...(Array.isArray(officialConfig.resolver.blockList) ? officialConfig.resolver.blockList : [officialConfig.resolver.blockList]),
        ...options.extMetroConfig.blockRegexp || [],
      ],
      useWatchman,
    },

    serializer: {
      processModuleFilter: filterModuleHOF({ projectRoot, SDKAlias, filterModules, filterModulesVersion, mode }),
      createModuleIdFactory,
      getModulesRunBeforeMainModule: (entryRelative: string) => {
        const officialRunBefore = officialConfig.serializer.getModulesRunBeforeMainModule(entryRelative)
        return [...officialRunBefore, ...extraModulesRunFirst]
      },
      getPolyfills: (opts: any) => {
        const officialPolyfills = officialConfig.serializer.getPolyfills(opts)
        return [...officialPolyfills, ...extraPolyfills]
      }
    },

    server: {
      port: (options as DevOptions).port || 8081,

      // 中间件的执行顺序: communityCLI => metroMiddleware.
      // 这个中间件中的逻辑, 会晚于 communityCLI 的中间件执行
      enhanceMiddleware: undefined,
    },

    transformer: {
      // babelPlugins: [],
      enableBabelRCLookup: false,
      getTransformOptions: async () => ({
        transform: {
          experimentalImportSupport: false,
          inlineRequires,
          unstable_disableES6Transforms: false,
        },
        preloadedModules: false,
        ramGroups: [],
      }),
      // asyncRequireModulePath: require.resolve('metro-runtime/src/modules/asyncRequire'),
      babelTransformerPath: path.join(__dirname, '../matrix-config/babelTransformer.js'),

      assetRegistryPath,
      // publicPath用来决定图片的地址, 开发环境需设置为空
      publicPath: '/',
    },

    cacheStores: [
      new FileStore({ root: CACHE_PATH })
    ],
    hasteMapCacheDirectory: CACHE_PATH,

    // 确保不同的小程序能产生不同的缓存值
    cacheVersion: projectConfig?.cacheVersion || '1',
    projectRoot,
    watchFolders,

    onReady: () => { },
    enableRootBabelRC: false,
  }

  return mergeConfig(
    officialDefaultConfig,
    officialCliConfig,
    hyextConfig,
    options.extMetroConfig,
  )
}
